library ieee;
use ieee.STD_LOGIC_1164.all;
use ieee.NUMERIC_STD.all;

use work.gencores_pkg.all;
use work.vme64x_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity serial_bridge_master is
  generic(
    g_clock_period : integer
    );
  port (

    afpga_rst_n_i : in std_logic;
    afpga_clk_i   : in std_logic;

    afpga_d_i     : in  std_logic_vector(15 downto 0);
    afpga_frame_i : in  std_logic;
    afpga_d_o     : out std_logic_vector(15 downto 0);
    afpga_frame_o : out std_logic;

    vme_berr_n_i : in std_logic;
    vme_ga_i : in std_logic_vector(4 downto 0);
    
    mem_req_i : in std_logic;
    mem_is_blt_i : in std_logic;
    mem_ack_o : out std_logic;
    mem_err_o : out std_logic;
    mem_we_i : in std_logic;
    mem_addr_i : in std_logic_vector(31 downto 0);
    mem_data_o : out std_logic_vector(31 downto 0);
    mem_data_i : in std_logic_vector(31 downto 0);
    mem_sel_i : in std_logic_vector(3 downto 0);
    
    
    -- Function decoder
    addr_decoder_i  : in  std_logic_vector(31 downto 1);
    addr_decoder_o  : out std_logic_vector(31 downto 1);
    decode_start_i  : in std_logic;
    decode_done_o   : out std_logic;
    am_i            : in std_logic_vector( 5 downto 0);
    decode_sel_o   : out  std_logic;

    -- CR/CSR space signals:
    cr_csr_addr_i   : in std_logic_vector(18 downto 2);
    cr_csr_data_i   : in  std_logic_vector( 7 downto 0);
    cr_csr_data_o   : out std_logic_vector( 7 downto 0);
    cr_csr_we_i     : in std_logic;
    cr_csr_req_i    : in std_logic;
    cr_csr_done_o   : out std_logic;
    
    
    module_enable_o : out  std_logic;
    bar_o           : out  std_logic_vector( 4 downto 0);

    -- Interrupts
    int_level_o     : out  std_logic_vector( 2 downto 0);
    int_vector_o    : out  std_logic_vector( 7 downto 0);
    irq_pending_o   : out  std_logic;
    irq_ack_i       : in std_logic;

    heartbeat_tx_o : out std_logic;
    heartbeat_rx_o : out std_logic
    );

end serial_bridge_master;

architecture rtl of serial_bridge_master is

  constant c_tag_csr_req     : std_logic_vector(3 downto 0) := x"0";
  constant c_tag_csr_rsp     : std_logic_vector(3 downto 0) := x"1";
  constant c_tag_config0     : std_logic_vector(3 downto 0) := x"2";
  constant c_tag_decode_req    : std_logic_vector(3 downto 0) := x"3";
  constant c_tag_decode_rsp : std_logic_vector(3 downto 0) := x"4";
  constant c_tag_config1     : std_logic_vector(3 downto 0) := x"5";
  constant c_tag_write_req     : std_logic_vector(3 downto 0) := x"6";
  constant c_tag_read_req     : std_logic_vector(3 downto 0) := x"7";
  constant c_tag_write_rsp     : std_logic_vector(3 downto 0) := x"8";
  constant c_tag_read_rsp     : std_logic_vector(3 downto 0) := x"9";

  signal afpga_dout      : std_logic_vector(15 downto 0);
  signal afpga_din       : std_logic_vector(15 downto 0);
  signal afpga_frame_out : std_logic;
  signal afpga_frame_in  : std_logic;

  signal mem_ack_int : std_logic;

  type t_tx_state is
    (
      IDLE,
      TX_CSR_REQ,
      TX_DECODE_REQ0,
      TX_DECODE_REQ1,
      RX_CSR_ACK,
      RX_DECODE_RSP,
      RX_DECODE_RSP1,
      RX_DECODE_RSP2,
      RX_MEM_WR_ACK,
      TX_MEM_WR_REQ0,
      TX_MEM_WR_REQ1,
      TX_MEM_WR_REQ2,
      TX_MEM_WR_REQ3,
      TX_MEM_RD_REQ0,
      TX_MEM_RD_REQ1,
      RX_MEM_RD_CPL0,
      RX_MEM_RD_CPL1,
      RX_MEM_RD_CPL2



      );

  type t_rx_state is
    (
      IDLE,
      RX_CONF0
      );
  
  signal tx_state : t_tx_state;
  signal rx_state : t_rx_state;

  signal heartbeat_cnt : unsigned(23 downto 0);
  signal heartbeat : std_logic;
begin
  mem_ack_o <= mem_ack_int;

  heartbeat_tx_o <= heartbeat;

  process(afpga_clk_i)
  begin
    if rising_edge(afpga_clk_i) then
      if afpga_rst_n_i = '0' then
        heartbeat_cnt <= (others => '0');
      else
        heartbeat_cnt <= heartbeat_cnt + 1;
        heartbeat <= std_logic(heartbeat_cnt(heartbeat_cnt'length - 1 ) );
      end if;
    end if;
  end process;
    
  
  afpga_d_o <= afpga_dout;
  afpga_frame_o <= afpga_frame_out;
  afpga_frame_in <= afpga_frame_i;
  afpga_din <= afpga_d_i;
  
  p_in_fsm : process(afpga_clk_i)
  begin
    if rising_edge(afpga_clk_i) then
      if afpga_rst_n_i = '0' then
        rx_state <= IDLE;
      else
        case rx_state is
          when IDLE =>
            if afpga_frame_in = '1' then
              case afpga_din(15 downto 12) is
                when c_tag_config0 =>
                  module_enable_o <= afpga_din(11);
                  bar_o <= afpga_din(10 downto 6);
                  heartbeat_rx_o <= afpga_din(5);
                when c_tag_config1 =>
                  int_level_o     <= afpga_din(2 downto 0);
                  int_vector_o    <=afpga_din(15 downto 8);
                  irq_pending_o   <= afpga_din(3);
                when others => null;
              end case;
            end if;
            when others => null;
        end case;
      end if;
    end if;
  end process;
  
  

  p_out_fsm : process(afpga_clk_i)
  begin
    if rising_edge(afpga_clk_i) then
      if afpga_rst_n_i = '0' then
        tx_state <= IDLE;
        afpga_frame_out <= '0';
        afpga_dout <= (others => '0');
        decode_done_o <= '0';
        decode_sel_o <= '0';
        cr_csr_done_o <= '0';
        mem_ack_int <= '0';
      else
        
        case tx_state is


          
          when IDLE =>
            cr_csr_done_o <= '0';
            afpga_frame_out <= '0';
            decode_done_o <= '0';
            mem_ack_int<= '0';
            
            if cr_csr_req_i = '1' then
              afpga_frame_out <= '1';
              afpga_dout <= c_tag_csr_req & cr_csr_we_i & cr_csr_addr_i(18 downto 8);
              tx_state <= TX_CSR_REQ;
            elsif decode_start_i = '1' then
              afpga_frame_out <= '1';
              afpga_dout(15 downto 12) <= c_tag_decode_req;
              afpga_dout(11 downto 6) <= am_i;
              tx_state <= TX_DECODE_REQ0;
            elsif mem_req_i = '1' and mem_ack_int = '0' then
              afpga_frame_out <= '1';
              afpga_dout(11 downto 8) <= mem_sel_i;
              afpga_dout(7) <= mem_is_blt_i;

              if mem_we_i = '1' then
                afpga_dout(15 downto 12) <= c_tag_write_req;
                tx_state <= TX_MEM_WR_REQ0;
              else
                afpga_dout(15 downto 12) <= c_tag_read_req;
                tx_state <= TX_MEM_RD_REQ0;
              end if;
              
            else
              afpga_frame_out <= '1';
              afpga_dout(15 downto 12) <= c_tag_config0;
              afpga_dout(11) <= vme_berr_n_i;
              afpga_dout(10 downto 6) <= vme_ga_i;
              afpga_dout(5) <= heartbeat;
              afpga_dout(4 downto 0) <= (others => '0');
            end if;
            


          when TX_MEM_WR_REQ0 =>
            afpga_frame_out <= '0';
            afpga_dout <= mem_addr_i(31 downto 16);
            tx_state <= TX_MEM_WR_REQ1;
          when TX_MEM_WR_REQ1 =>
            afpga_frame_out <= '0';
            afpga_dout <= mem_addr_i(15 downto 0);
            tx_state <= TX_MEM_WR_REQ2;

          when TX_MEM_RD_REQ0 =>
            afpga_frame_out <= '0';
            afpga_dout <= mem_addr_i(31 downto 16);
            tx_state <= TX_MEM_RD_REQ1;
          when TX_MEM_RD_REQ1 =>
            afpga_frame_out <= '0';
            afpga_dout <= mem_addr_i(15 downto 0);
            tx_state <= RX_MEM_RD_CPL0;
            
          when TX_MEM_WR_REQ2 =>
            afpga_frame_out <= '0';
            afpga_dout <= mem_data_i(31 downto 16);
            tx_state <= TX_MEM_WR_REQ3;
          when TX_MEM_WR_REQ3 =>
            afpga_frame_out <= '0';
            afpga_dout <= mem_data_i(15 downto 0);
            tx_state <= RX_MEM_WR_ACK;

          when RX_MEM_WR_ACK =>
            if afpga_frame_in = '1' and afpga_din(15 downto 12) = c_tag_write_rsp then
              mem_ack_int <= '1';
              mem_err_o <= afpga_din(11);
              tx_state <= IDLE;
            end if;

          when RX_MEM_RD_CPL0 =>
            if afpga_frame_in = '1' and afpga_din(15 downto 12) = c_tag_read_rsp then
              mem_err_o <= afpga_din(11);
              tx_state <= RX_MEM_RD_CPL1;
            end if;

          when RX_MEM_RD_CPL1 =>
            mem_data_o(31 downto 16) <= afpga_din;
            tx_state <= RX_MEM_RD_CPL2;

          when RX_MEM_RD_CPL2 =>
            mem_data_o(15 downto 0) <= afpga_din;
            mem_ack_int <= '1';
            tx_state <= IDLE;
            
            
          when TX_CSR_REQ =>
            afpga_frame_out <= '0';
            afpga_dout <= cr_csr_addr_i(7 downto 2) & "00" & cr_csr_data_i;
            tx_state <= RX_CSR_ACK;

          when RX_CSR_ACK =>
            if afpga_frame_in = '1' and afpga_din(15 downto 12) = c_tag_csr_rsp then
              cr_csr_done_o <= '1';
              cr_csr_data_o <= afpga_din(7 downto 0);
              tx_state <= IDLE;
            end if;

          when TX_DECODE_REQ0 =>
            afpga_dout <= addr_decoder_i(31 downto 16);
            afpga_frame_out <= '0';
            tx_state <= TX_DECODE_REQ1;

          when TX_DECODE_REQ1 =>
            afpga_dout <= addr_decoder_i(15 downto 1) & '0';
            tx_state <= RX_DECODE_RSP;

            
          when RX_DECODE_RSP =>
            if afpga_frame_in = '1' and afpga_din(15 downto 12) = c_tag_decode_rsp then
              tx_state <= RX_DECODE_RSP1;
              decode_sel_o <= afpga_din(11);
            end if;

          when RX_DECODE_RSP1 =>
            addr_decoder_o(31 downto 16) <= afpga_din;
            tx_state <= RX_DECODE_RSP2;
          when RX_DECODE_RSP2 =>
            addr_decoder_o(15 downto 1) <= afpga_din(15 downto 1);
            decode_done_o <= '1';
            tx_state <= IDLE;
              
            
        end case;


      end if;

    end if;

  end process;


end rtl;
