--------------------------------------------------------------------------------
-- CERN (BE-CO-HT)
-- VME64x Core
-- http://www.ohwr.org/projects/vme64x-core
--------------------------------------------------------------------------------
--
-- unit name:     VME_bus
--
-- description:
--
--   This block acts as interface between the VMEbus and the CR/CSR space or
--   WB bus.
--
--------------------------------------------------------------------------------
-- GNU LESSER GENERAL PUBLIC LICENSE
--------------------------------------------------------------------------------
-- This source file is free software; you can redistribute it and/or modify it
-- under the terms of the GNU Lesser General Public License as published by the
-- Free Software Foundation; either version 2.1 of the License, or (at your
-- option) any later version. This source is distributed in the hope that it
-- will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
-- of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
-- See the GNU Lesser General Public License for more details. You should have
-- received a copy of the GNU Lesser General Public License along with this
-- source; if not, download it from http://www.gnu.org/licenses/lgpl-2.1.html
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.vme64x_pkg.all;
use work.wishbone_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity serial_bridge_slave is
  generic (
    g_CLOCK_PERIOD  : integer
  );
  port (

    sfpga_rst_n_i : in std_logic;
    sfpga_clk_i   : in std_logic;

    sfpga_d_i     : in  std_logic_vector(15 downto 0);
    sfpga_frame_i : in  std_logic;
    sfpga_d_o     : out std_logic_vector(15 downto 0);
    sfpga_frame_o : out std_logic;

    vme_aux_valid_o : out std_logic;
    vme_berr_n_o : out std_logic;
    vme_ga_o : out std_logic_vector(4 downto 0);
    
    -- WB signals
    mem_req_o : out std_logic;
    mem_is_blt_o : out std_logic;
    mem_ack_i : in std_logic;
    mem_err_i : in std_logic;
    mem_we_o : out std_logic;
    mem_addr_o : out std_logic_vector(31 downto 0);
    mem_data_o : out std_logic_vector(31 downto 0);
    mem_data_i : in std_logic_vector(31 downto 0);
    mem_sel_o : out std_logic_vector(3 downto 0);
    
    
    -- Function decoder
    addr_decoder_i  : in  std_logic_vector(31 downto 1);
    addr_decoder_o  : out std_logic_vector(31 downto 1);
    decode_start_o  : out std_logic;
    decode_done_i   : in std_logic;
    am_o            : out std_logic_vector( 5 downto 0);
    decode_sel_i    : in  std_logic;

    -- CR/CSR space signals:
    cr_csr_addr_o   : out std_logic_vector(18 downto 2);
    cr_csr_data_i   : in  std_logic_vector( 7 downto 0);
    cr_csr_data_o   : out std_logic_vector( 7 downto 0);
    cr_csr_we_o     : out std_logic;
    cr_csr_req_o    : out std_logic;
    cr_csr_done_i : in std_logic;

    
    
    module_enable_i : in  std_logic;
    bar_i           : in  std_logic_vector( 4 downto 0);

    irq_i : in std_logic := '0';

    -- Interrupts
    int_level_i     : in  std_logic_vector( 2 downto 0);
    int_vector_i    : in  std_logic_vector( 7 downto 0);
    irq_pending_i   : in  std_logic;
    irq_ack_o       : out std_logic;

    heartbeat_tx_o : out std_logic;
    heartbeat_rx_o : out std_logic

  );
end serial_bridge_slave;

architecture rtl of serial_bridge_slave is
  
  constant c_tag_csr_req     : std_logic_vector(3 downto 0) := x"0";
  constant c_tag_csr_rsp     : std_logic_vector(3 downto 0) := x"1";
  constant c_tag_config0     : std_logic_vector(3 downto 0) := x"2";
  constant c_tag_decode_req    : std_logic_vector(3 downto 0) := x"3";
  constant c_tag_decode_rsp : std_logic_vector(3 downto 0) := x"4";
  constant c_tag_config1     : std_logic_vector(3 downto 0) := x"5";
  constant c_tag_write_req     : std_logic_vector(3 downto 0) := x"6";
  constant c_tag_read_req     : std_logic_vector(3 downto 0) := x"7";
  constant c_tag_write_rsp     : std_logic_vector(3 downto 0) := x"8";
  constant c_tag_read_rsp     : std_logic_vector(3 downto 0) := x"9";

  signal sfpga_dout      : std_logic_vector(15 downto 0);
  signal sfpga_din       : std_logic_vector(15 downto 0);
  signal sfpga_frame_out : std_logic;
  signal sfpga_frame_in  : std_logic;
  signal sfpga_clk_n     : std_logic;

  signal cr_csr_write : std_logic;

  signal addr_decoder_reg : std_logic_vector(31 downto 1);
  
  type t_rx_state is
    (
      IDLE,
      RX_CSR_REQ0,
      TX_CSR_ACK,
      TX_CSR_WAIT1,
      TX_CSR_WAIT2,
      RX_DECODE_REQ0,
      RX_DECODE_REQ1,
      TX_DECODE_RSP,
      TX_DECODE_RSP1,
      TX_DECODE_RSP2,
      RX_MEM_WR_REQ0,
      RX_MEM_WR_REQ1,
      RX_MEM_WR_REQ2,
      RX_MEM_WR_REQ3,
      RX_MEM_RD_REQ0,
      RX_MEM_RD_REQ1,
      RX_MEM_RD_WAIT_ACK,
      RX_MEM_WR_WAIT_ACK,
      TX_MEM_RD_CPL0,
      TX_MEM_RD_CPL1
      );

  
  signal rx_state : t_rx_state;

  signal conf_cnt : unsigned(1 downto 0);
  signal mem_data_reg : std_logic_vector(31 downto 0);

  signal heartbeat_cnt : unsigned(23 downto 0);
  signal heartbeat : std_logic;

  
begin

  heartbeat_tx_o <= heartbeat;

  process(sfpga_clk_i)
  begin
    if rising_edge(sfpga_clk_i) then
      if sfpga_rst_n_i = '0' then
        heartbeat_cnt <= (others => '0');
      else
        heartbeat_cnt <= heartbeat_cnt + 1;
        heartbeat <= std_logic(heartbeat_cnt(heartbeat_cnt'length - 1 ) );
      end if;
    end if;
  end process;
  


  sfpga_frame_in <= sfpga_frame_i;
  sfpga_din <= sfpga_d_i;

  sfpga_frame_o <= sfpga_frame_out;
  sfpga_d_o <= sfpga_dout;

    p_in_fsm : process(sfpga_clk_i)
    variable send_cfg : boolean := false;
  begin
    if rising_edge(sfpga_clk_i) then
      if sfpga_rst_n_i = '0' then
        rx_state <= IDLE;
        vme_aux_valid_o <= '0';
        conf_cnt <= (others => '0');
        decode_start_o <= '0';
        cr_csr_req_o <= '0';
        cr_csr_we_o <= '0';
        mem_req_o <= '0';

        sfpga_dout <= (others => '0');
        sfpga_frame_out <= '0';
        
      else

       -- report "Din "  & integer'image(to_integer(unsigned(sfpga_din)));
        
        case rx_state is
          
          when IDLE =>
            decode_start_o <= '0';
            cr_csr_req_o <= '0';
            cr_csr_we_o <= '0';
            sfpga_frame_out <= '0';
            send_cfg := false;
            mem_req_o <= '0';
            
            
            if sfpga_frame_in = '1' then
              -- report "Din2 "  & integer'image(to_integer(unsigned(sfpga_din)));
              
              -- report "got frame" & integer'image(to_integer(unsigned(sfpga_din)));
              
              if sfpga_din(15 downto 12) = c_tag_decode_req then
--                report "got decode req";
                am_o <= sfpga_din(11 downto 6);
                rx_state <= RX_DECODE_REQ0;
                send_cfg := false;
              elsif sfpga_din(15 downto 12) = c_tag_config0 then
                --              report "decode conf0 req";
                vme_berr_n_o <= sfpga_din(11);
                vme_ga_o <= sfpga_din(10 downto 6);
                heartbeat_rx_o <= sfpga_din(5);
                vme_aux_valid_o <= '1';
                send_cfg := true;
              elsif sfpga_din(15 downto 12) = c_tag_csr_req then
                cr_csr_write <= sfpga_din(11);
                cr_csr_addr_o(18 downto 8) <= sfpga_din(10 downto 0);
                rx_state <= RX_CSR_REQ0;
                send_cfg := false;
              elsif sfpga_din(15 downto 12) = c_tag_write_req then
                rx_state <= RX_MEM_WR_REQ0;
                mem_we_o <= '1';
                mem_sel_o <= sfpga_din(11 downto 8);
                mem_is_blt_o <= sfpga_din(7);
              elsif sfpga_din(15 downto 12) = c_tag_read_req then
                rx_state <= RX_MEM_RD_REQ0;
                mem_we_o <= '0';
                mem_sel_o <= sfpga_din(11 downto 8);
                mem_is_blt_o <= sfpga_din(7);
                
              else
                send_cfg := true;
              end if;
              
              
            else
              send_cfg := true;
            end if;
            

            if send_cfg then
              conf_cnt <= conf_cnt + 1;

              if conf_cnt = 0 then
                sfpga_frame_out <= '1';
                sfpga_dout(15 downto 12) <= c_tag_config0;
                sfpga_dout(11) <= module_enable_i;
                sfpga_dout(10 downto 6) <= bar_i;
                sfpga_dout(5) <= heartbeat;
              elsif conf_cnt = 1 then
                sfpga_frame_out <= '1';

                sfpga_dout(2 downto 0) <= int_level_i;
                sfpga_dout(15 downto 8) <= int_vector_i;
                sfpga_dout(3) <= irq_pending_i;

                
                
                
              end if;
              
            end if;

          when RX_MEM_WR_REQ0 =>
            mem_addr_o(31 downto 16) <= sfpga_din;
            rx_state <= RX_MEM_WR_REQ1;
          when RX_MEM_WR_REQ1 =>
            mem_addr_o(15 downto 0) <= sfpga_din;
            rx_state <= RX_MEM_WR_REQ2;
          when RX_MEM_WR_REQ2 =>
            mem_data_o(31 downto 16) <= sfpga_din;
            rx_state <= RX_MEM_WR_REQ3;
          when RX_MEM_WR_REQ3 =>
            mem_data_o(15 downto 0) <= sfpga_din;
            mem_req_o <= '1';
            rx_state <= RX_MEM_WR_WAIT_ACK;
          when RX_MEM_WR_WAIT_ACK =>
            if mem_ack_i = '1' then
              mem_req_o <= '0';
              sfpga_frame_out <= '1';
              sfpga_dout(15 downto 12) <= c_tag_write_rsp;
              sfpga_dout(11) <= mem_err_i;
              rx_state <= IDLE;
            end if;
            
            

          when RX_MEM_RD_REQ0 =>
            mem_addr_o(31 downto 16) <= sfpga_din;
            rx_state <= RX_MEM_RD_REQ1;
          when RX_MEM_RD_REQ1 =>
            mem_addr_o(15 downto 0) <= sfpga_din;
            rx_state <= RX_MEM_RD_WAIT_ACK;
            mem_req_o <= '1';
          when RX_MEM_RD_WAIT_ACK =>
            if mem_ack_i = '1' then
              mem_data_reg <= mem_data_i;
              sfpga_frame_out <= '1';
              sfpga_dout(15 downto 12) <= c_tag_read_rsp;
              sfpga_dout(11) <= mem_err_i;
              mem_req_o <= '0';
              rx_state <= TX_MEM_RD_CPL0;
            end if;

          when TX_MEM_RD_CPL0 =>
            sfpga_dout <= mem_data_reg(31 downto 16);
            rx_state <= TX_MEM_RD_CPL1;

          when TX_MEM_RD_CPL1 =>
            sfpga_dout <= mem_data_reg(15 downto 0);
            rx_state <= IDLE;
            
          when RX_CSR_REQ0 =>
            rx_state <= TX_CSR_WAIT1;
            cr_csr_we_o <= '0';
            
            
            cr_csr_addr_o(7 downto 2) <= sfpga_din(15 downto 10);
            cr_csr_data_o(7 downto 0) <= sfpga_din(7 downto 0);
            cr_csr_we_o <= cr_csr_write;

          when TX_CSR_WAIT1 =>
            rx_state <= TX_CSR_WAIT2;
          when TX_CSR_WAIT2 =>
            rx_state <= TX_CSR_ACK;
            
          when TX_CSR_ACK =>
            cr_csr_we_o <= '0';

            sfpga_frame_out <= '1';
            sfpga_dout(15 downto 12) <= c_tag_csr_rsp;
            sfpga_dout(7 downto 0) <= cr_csr_data_i;

            rx_state <= IDLE;

            
          when RX_DECODE_REQ0 =>
            addr_decoder_o(31 downto 16) <= sfpga_din;
            rx_state <= RX_DECODE_REQ1;
          when RX_DECODE_REQ1 =>
            addr_decoder_o(15 downto 1) <= sfpga_din(15 downto 1);
            decode_start_o <= '1';
            rx_state <= TX_DECODE_RSP;
          when TX_DECODE_RSP =>
            decode_start_o <= '0';
            if decode_done_i = '1' then
              sfpga_frame_out <= '1';
              sfpga_dout(15 downto 12) <= c_tag_decode_rsp;
              sfpga_dout(11) <= decode_sel_i;
              rx_state <= TX_DECODE_RSP1;
              addr_decoder_reg <= addr_decoder_i;
            end if;

          when TX_DECODE_RSP1 =>
            sfpga_frame_out <= '0';
            sfpga_dout <= addr_decoder_reg(31 downto 16);
            rx_state <= TX_DECODE_RSP2;

          when TX_DECODE_RSP2 =>
            sfpga_frame_out <= '0';
            sfpga_dout(15 downto 1) <= addr_decoder_reg(15 downto 1);
            rx_state <= IDLE;
            
            
          when others => null;
        end case;
      end if;
    end if;
  end process;

  
end rtl;
