library ieee;
use ieee.STD_LOGIC_1164.all;
use ieee.NUMERIC_STD.all;

use work.gencores_pkg.all;
use work.vme64x_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity bridge_master is
  generic(
    g_clock_period : integer
    );
  port (

    afpga_rst_n_i : in std_logic;
    afpga_clk_i   : in std_logic;

    afpga_d_i     : in  std_logic_vector(7 downto 0);
    afpga_frame_i : in  std_logic;
    afpga_d_o     : out std_logic_vector(7 downto 0);
    afpga_frame_o : out std_logic;

    mem_req_i : in std_logic;
    mem_is_blt_i : in std_logic;
    mem_ack_o : out std_logic;
    mem_err_o : out std_logic;
    mem_we_i : in std_logic;
    mem_addr_i : in std_logic_vector(31 downto 0);
    mem_data_o : out std_logic_vector(31 downto 0);
    mem_data_i : in std_logic_vector(31 downto 0);
    mem_sel_i : in std_logic_vector(3 downto 0);
    
    
    -- Function decoder
    addr_decoder_i  : in  std_logic_vector(31 downto 1);
    addr_decoder_o  : out std_logic_vector(31 downto 1);
    decode_start_i  : in std_logic;
    decode_done_o   : out std_logic;
    am_i            : in std_logic_vector( 5 downto 0);
    decode_sel_o   : out  std_logic;

    -- CR/CSR space signals:
    cr_csr_addr_i   : in std_logic_vector(18 downto 2);
    cr_csr_data_i   : in  std_logic_vector( 7 downto 0);
    cr_csr_data_o   : out std_logic_vector( 7 downto 0);
    cr_csr_we_i     : in std_logic;
    cr_csr_req_i    : in std_logic;
1    cr_csr_done_o   : out std_logic;
    
    
    module_enable_o : out  std_logic;
    bar_o           : out  std_logic_vector( 4 downto 0);

    -- Interrupts
    int_level_o     : out  std_logic_vector( 2 downto 0);
    int_vector_o    : out  std_logic_vector( 7 downto 0);
    irq_pending_o   : out  std_logic;
    irq_ack_i       : in std_logic
    );

end bridge_master;

architecture rtl of bridge_master is

  constant c_tag_csr_req     : std_logic_vector(3 downto 0) := x"0";
  constant c_tag_csr_rsp     : std_logic_vector(3 downto 0) := x"1";
  constant c_tag_config     : std_logic_vector(3 downto 0) := x"2";
  constant c_tag_decode_req    : std_logic_vector(3 downto 0) := x"2";
  constant c_tag_read_cpl : std_logic_vector(3 downto 0) := x"3";
  constant c_tag_irq      : std_logic_vector(3 downto 0) := x"4";
  constant c_tag_iackin   : std_logic_vector(3 downto 0) := x"5";
  constant c_tag_iackout  : std_logic_vector(3 downto 0) := x"6";
  constant c_tag_aux      : std_logic_vector(3 downto 0) := x"7";

  signal afpga_dout      : std_logic_vector(15 downto 0);
  signal afpga_din       : std_logic_vector(15 downto 0);
  signal afpga_frame_out : std_logic;
  signal afpga_frame_in  : std_logic;
  signal afpga_clk_n     : std_logic;


  type t_state is
    (
      IDLE, TX_CSR_REQ,
      RX_CSR_ACK );

  type t_rx_state is
    (
      IDLE,
      RX_CONF0
      );
  
  signal state : t_state;
  signal rx_state : t_state;
  
begin

  afpga_clk_n <= not afpga_clk_i;

  gen_ddr_ios : for i in 0 to 7 generate

    U_ODDR_Data : ODDR2
      generic map (
        DDR_ALIGNMENT => "C0")
      port map (
        Q  => afpga_d_o(i),
        C0 => afpga_clk_i,
        C1 => afpga_clk_n,
        CE => '1',
        D0 => afpga_dout(2*i),
        D1 => afpga_dout(2*i+1),
        R  => '0',
        S  => '0');

    U_IDDR_Data : IDDR2
      generic map (
        DDR_ALIGNMENT => "C0")
      port map (
        Q0 => afpga_din(2*i+1),
        Q1 => afpga_din(2*i),
        C0 => afpga_clk_i,
        C1 => afpga_clk_n,
        CE => '1',
        D  => afpga_d_i(i),
        R  => '0',
        S  => '0');

  end generate gen_ddr_ios;

  U_IDDR_Frame : IDDR2
    generic map (
      DDR_ALIGNMENT => "C0")
    port map (
      Q0 => afpga_frame_in,
      Q1 => open,
      C0 => afpga_clk_i,
      C1 => afpga_clk_n,
      CE => '1',
      D  => afpga_frame_i,
      R  => '0',
      S  => '0');

  U_ODDR_Frame : ODDR2
    generic map (
      DDR_ALIGNMENT => "C0")
    port map (
      Q  => afpga_frame_o,
      C0 => afpga_clk_i,
      C1 => afpga_clk_n,
      CE => '0',
      D0 => afpga_frame_out,
      D1 => afpga_frame_out,
      R  => '0',
      S  => '0');

  

  p_in_fsm : process(afpga_clk_i)
  begin
    if rising_edge(afpga_clk_i) then
      if afpga_rst_n_i = '0' then
        rx_state <= IDLE;
      else
        case rx_state is
          when IDLE =>
            if afpga_frame_in = '1' then
              case afpga_din(15 downto 12) is
                when c_tag_config =>

                  module_enable_o <= afpga_din(11);
                  bar_o <= afpga_din(10 downto 6);
                  rx_state <= RX_CONF0;
                when others => null;
              end case;
            end if;
          when RX_CONF0 =>
            int_level_o     <= afpga_din(2 downto 0);
            int_vector_o    <=afpga_din(15 downto 8);
            irq_pending_o   <= afpga_din(3);
            rx_state <= IDLE;
        end case;
      end if;
    end if;
  end process;
  
  

  p_out_fsm : process(afpga_clk_i)
  begin
    if rising_edge(afpga_clk_i) then
      if afpga_rst_n_i = '0' then
        state <= IDLE;
        afpga_frame_out <= '0';
        
      else
        
        case state is
          when IDLE =>
            cr_csr_done_o <= '0';
            afpga_frame_out <= '0';
            
            if cr_csr_req_i = '1' then
              afpga_frame_out <= '1';
              afpga_dout <= c_tag_csr_req & cr_csr_we_i & cr_csr_addr_i(18 downto 5);
              state <= TX_CSR_REQ;
            end if;

            if decode_start_i = '1' then
              afpga_frame_out <= '1';
              afpga_dout <= c_tag_decode_req & am_i & "000000";
              state <= TX_DECODE_REQ0;
            end if;
            
            
          when TX_CSR_REQ =>
            afpga_frame_out <= '0';
            afpga_dout <= cr_csr_addr_i(4 downto 0) & "000" & cr_csr_data_i;
            state <= RX_CSR_ACK;

          when RX_CSR_ACK =>
            if afpga_frame_in = '1' and afpga_din(15 downto 12) = c_tag_csr_rsp then
              cr_csr_done_o <= '1';
              cr_csr_data_o <= afpga_din(7 downto 0);
              state <= IDLE;
            end if;

          when TX_DECODE_REQ0 =>
            afpga_dout <= addr_decoder_i(31 downto 16);
            state <= TX_DECODE_REQ1;

          when TX_DECODE_REQ1 =>
            afpga_dout <= addr_decoder_i(15 downto 1) & '0';
            state <= TX_DECODE_REQ1;

            
          when RX_DECODE_RSP =>
            
            
            
        end case;


      end if;

    end if;

  end process;


end rtl;
